@@ -1,8 +1,9 @@
+import hashlib
 import json
 from os import getenv
-from typing import Any, Dict, Optional, List, Callable, Coroutine
+from pydantic import BaseModel
+from typing import Any, Dict, Optional, List, Callable
 from xpander_sdk import (
-    GraphItem,
     UserDetails,
     Memory,
     XpanderClient,
@@ -29,6 +30,12 @@ from chainlit.step import StepDict
 from chainlit.element import ElementDict
 
 AGENT_END_TOOL = "xpfinish-agent-execution-finished"
+MAX_TOOL_CALLS = 15
+MAX_SAME_TOOL_IN_A_ROW = 5
+
+class ChainlitToolCallWithLLMResponse(BaseModel):
+    tool_call: Any
+    completion_response: Any = None
 
 class ChainlitXpanderDataLayer(BaseDataLayer):
     """
@@ -296,7 +303,7 @@ class ChainlitXpanderDataLayer(BaseDataLayer):
             tags (Optional[List[str]]): List of tags.
         """
         author: dict = self.thread_authors_cache.get(thread_id, None)
-        if not author:
+        if not author or author == None:
             pass
 
         xpander = XpanderClient(
@@ -346,7 +353,8 @@ class ChainlitAdapter(SDKAdapter):
         organization_id: Optional[str] = None,
         user: Optional[User] = None,
         with_agent_end_tool: Optional[bool] = False,
-        should_reset_cache: Optional[bool] = False
+        should_reset_cache: Optional[bool] = False,
+        with_metrics_report: Optional[bool] = False
     ):
         """
         Initialize the ChainlitAdapter.
@@ -356,13 +364,16 @@ class ChainlitAdapter(SDKAdapter):
             agent_id (str): The ID of the xpander.ai agent.
             base_url (Optional[str], optional): The base URL for the xpander.ai API. Defaults to None.
             organization_id (Optional[str], optional): The organization ID, if applicable. Defaults to None.
+            with_metrics_report (Optional[bool], optional): If to auto-report metrics (llm & execution). Default to False.
         """
-        super().__init__(api_key, agent_id, base_url, organization_id, should_reset_cache)
+        super().__init__(api_key=api_key, agent_id=agent_id, base_url=base_url, organization_id=organization_id, should_reset_cache=should_reset_cache, with_metrics_report=with_metrics_report)
         
         if not with_agent_end_tool:
             self.agent.disable_agent_end_tool()  # No need since SmolAgents handles it.
         
         self.named_threads = set()
+        self.execution_tool_calls: Dict[str, list[str]] = {}
+        
         if user:
             self.agent.update_user_details(
                 user_details=UserDetails(
@@ -471,56 +482,80 @@ class ChainlitAdapter(SDKAdapter):
         self,
         tool_calls: Optional[Dict[int, ToolCall]] = None,
         tool_call_requests: Optional[List[Any]] = None,
-        completion_response: Any
+        completion_response: Any = None
 
-    ) -> Dict[int, ToolCall]:
+    ) -> Dict[int, ChainlitToolCallWithLLMResponse]:
         """
         Aggregate tool calls from tool call requests.
 
         Args:
-            tool_calls (Optional[Dict[int, ToolCall]], optional): Existing tool calls. Defaults to None.
+            tool_calls (Optional[Dict[int, ChainlitToolCallWithLLMResponse]], optional): Existing tool calls. Defaults to None.
             tool_call_requests (Optional[List[Any]], optional): List of tool call requests. Defaults to None.
 
         Returns:
-            Dict[int, ToolCall]: Aggregated tool calls.
+            Dict[int, ChainlitToolCallWithLLMResponse]: Aggregated tool calls.
         """
         if not tool_calls:
-            tool_calls = {}
+            tool_calls: Dict[int, ChainlitToolCallWithLLMResponse] = {}
 
         if tool_call_requests:
             for tc in tool_call_requests:
                 if tc.index not in tool_calls:
-                    tool_calls[tc.index] = ToolCall(
-                        name=tc.function.name,
-                        tool_call_id=tc.id,
-                        type=ToolCallType.XPANDER if not tc.function.name.startswith("xpLocal") else ToolCallType.LOCAL,
-                        payload="",
+                    tool_calls[tc.index] = ChainlitToolCallWithLLMResponse(
+                        tool_call=ToolCall(
+                            name=tc.function.name,
+                            tool_call_id=tc.id,
+                            type=ToolCallType.XPANDER if not tc.function.name.startswith("xpLocal") else ToolCallType.LOCAL,
+                            payload="",
+                        ),
+                        completion_response=completion_response
                     )
+                   
                 else:
-                    tool_calls[tc.index].payload += tc.function.arguments
+                    tool_calls[tc.index].tool_call.payload += tc.function.arguments
+                    completion_response['choices'][0]['delta']['tool_calls'] = [{"function":{"arguments": tool_calls[tc.index].tool_call.payload}}]
+                    tool_calls[tc.index].completion_response = completion_response
         
         return tool_calls
 
     @cl.step(type="tool", name="xpander-ai")
-    async def process_tool_calls(self, tool_calls: Dict[int, ToolCall], on_tool_call: Optional[Callable], local_tools: Dict[str, Callable] = None) -> str | None:
+    async def process_tool_calls(self, tool_calls: Dict[int, ChainlitToolCallWithLLMResponse], on_node_enter: Optional[Callable], on_node_error: Optional[Callable], local_tools: Dict[str, Callable] = None) -> str | None:
         """
         Process tool calls by formatting their payloads and executing them.
 
         Args:
-            tool_calls (Dict[int, ToolCall]): The tool calls to process.
+            tool_calls (Dict[int, ChainlitToolCallWithLLMResponse]): The tool calls to process.
         """
         
         current_execution = self.agent.execution.to_dict()
         user_settings = cl.user_session.get("user_settings",{})
         is_debug_mode_active = user_settings.get("debug_mode", True)
+        is_repeating_tool_call = False
+        is_max_tool_call_reached = False
         
         is_sub_agent = True if "parentExecution" in current_execution and current_execution['parentExecution'] else False
         
-        for tc in tool_calls.values():
+        for tc_raw in tool_calls.values():
+            tc = tc_raw.tool_call
             if tc.payload:
                 tc.payload = json.loads(tc.payload)
 
-        tool_calls_list = list(tool_calls.values())
+        tool_calls_list: list[ToolCall] = [tc.tool_call for tc in tool_calls.values()]
+
+        # Count tool calls per execution
+        execution_id = self.agent.execution.id
+
+        if execution_id in self.execution_tool_calls:
+            # Check max tool call count
+            if len(self.execution_tool_calls[execution_id]) >= MAX_TOOL_CALLS:
+                is_max_tool_call_reached = True
+                
+            # Check if the last 3 tool calls are the same
+            if len(self.execution_tool_calls[execution_id]) >= MAX_SAME_TOOL_IN_A_ROW:
+                last_three_calls = self.execution_tool_calls[execution_id][-MAX_SAME_TOOL_IN_A_ROW:]
+                if len(set(last_three_calls)) == 1:  # All three are the same
+                    is_repeating_tool_call = True
+
 
         self.agent.add_messages(
             messages=[
@@ -535,11 +570,13 @@ class ChainlitAdapter(SDKAdapter):
                         }
                         for tc in tool_calls_list
                     ],
+                    "completion_response": tool_calls[0].completion_response
                 }
             ]
         )
         
-        for tool_call in tool_calls.values():
+        for raw_tool_call in tool_calls.values():
+            tool_call = raw_tool_call.tool_call
             is_agent_end_tool = True if tool_call.name == AGENT_END_TOOL else False
             
             current_step = cl.context.current_step
@@ -555,18 +592,18 @@ class ChainlitAdapter(SDKAdapter):
             if tool_agent_id:
                 sub_agent_node = self.agent.graph.find_node_by_item_id(item_id=tool_agent_id)
                 if sub_agent_node:
-                    if on_tool_call:
-                        await on_tool_call(tool_call=tool_call,graph_item=sub_agent_node)
+                    if on_node_enter:
+                        await on_node_enter(tool_call=tool_call,graph_item=sub_agent_node)
                     tool_name = sub_agent_node.name
             else:
                 # report to parent window
                 graph_item = self.agent.graph.find_node_by_item_id(item_id=tool_call.name)
-                if graph_item and on_tool_call:
-                    await on_tool_call(tool_call=tool_call,graph_item=graph_item)
+                if graph_item and on_node_enter:
+                    await on_node_enter(tool_call=tool_call,graph_item=graph_item)
                     
             if is_sub_agent:
                 if is_agent_end_tool:
-                    tool_name = f"{self.agent.name} - Final Thought"
+                    tool_name = f"{self.agent.name} - Final Thoughts"
                 else:
                     tool_name = f"{self.agent.name} - {tool_call.name}"
                     
@@ -604,7 +641,40 @@ class ChainlitAdapter(SDKAdapter):
                 # report result to the memory
                 self.agent.add_tool_call_results(tool_call_results=[tool_call_result])
             else:
-                tool_call_result = self.agent.run_tool(tool=tool_call, payload_extension={"headers":{"x-xpander-source-node-type": SourceNodeType.ASSISTANT.value.lower()}})
+                # failures
+                if is_max_tool_call_reached or is_repeating_tool_call:
+                    if is_max_tool_call_reached:
+                        error = f"Oops! You've reached the maximum allowed tool executions ({MAX_TOOL_CALLS}). Please adjust your input or try again later."
+                    elif is_repeating_tool_call:
+                        error = f"Oops! The last {MAX_SAME_TOOL_IN_A_ROW} tool calls were the same ({last_three_calls[0]}). The agent cannot proceed with this tool."
+                    
+                    tool_call_result = ToolCallResult(function_name=tool_call.name,tool_call_id=tool_call.tool_call_id,payload=tool_call.payload,result=error,is_success=False,is_error=True)
+                    self.agent.add_tool_call_results(tool_call_results=[tool_call_result])
+                    self.agent.add_messages(
+                        messages=[
+                            {
+                                "role": "system",
+                                "content": error
+                            }
+                        ]
+                    )
+                    
+                    cl.context.current_step.show_input = False
+                    cl.context.current_step.default_open = True
+                    cl.context.current_step.output = error
+                    await cl.context.current_step.update()
+                    raise Exception(error)
+                else: # execute the tool
+                    tool_call_result = self.agent.run_tool(tool=tool_call, payload_extension={"headers":{"x-xpander-source-node-type": SourceNodeType.ASSISTANT.value.lower()}})
+            
+            if not tool_call_result.is_success and is_debug_mode_active:
+                current_step.is_error = True
+                await on_node_error(tool_call_result=tool_call_result,graph_item=graph_item)
+            
+            if not execution_id in self.execution_tool_calls:
+                self.execution_tool_calls[execution_id] = [tool_call_result.function_name]
+            else:
+                self.execution_tool_calls[execution_id].append(tool_call_result.function_name)
             
             if is_debug_mode_active:
                 if is_agent_end_tool:
